#ifndef WaveFile_h
#define WaveFile_h
#include <SdFat.h>
#ifdef SD_FAT_VERSION
/**
Offset  Size  Name             Description

The canonical WAVE format starts with the RIFF header:

0         4   ChunkID          Contains the letters "RIFF" in ASCII form
                               (0x52494646 big-endian form).
4         4   ChunkSize        36 + SubChunk2Size, or more precisely:
                               4 + (8 + SubChunk1Size) + (8 + SubChunk2Size)
                               This is the size of the rest of the chunk 
                               following this number.  This is the size of the 
                               entire file in bytes minus 8 bytes for the
                               two fields not included in this count:
                               ChunkID and ChunkSize.
8         4   Format           Contains the letters "WAVE"
                               (0x57415645 big-endian form).

The "WAVE" format consists of two subchunks: "fmt " and "data":
The "fmt " subchunk describes the sound data's format:

12        4   Subchunk1ID      Contains the letters "fmt "
                               (0x666d7420 big-endian form).
16        4   Subchunk1Size    16 for PCM.  This is the size of the
                               rest of the Subchunk which follows this number.
20        2   AudioFormat      PCM = 1 (i.e. Linear quantization)
                               Values other than 1 indicate some 
                               form of compression.
22        2   NumChannels      Mono = 1, Stereo = 2, etc.
24        4   SampleRate       8000, 44100, etc.
28        4   ByteRate         == SampleRate * NumChannels * BitsPerSample/8
32        2   BlockAlign       == NumChannels * BitsPerSample/8
                               The number of bytes for one sample including
                               all channels. I wonder what happens when
                               this number isn't an integer?
34        2   BitsPerSample    8 bits = 8, 16 bits = 16, etc.
          2   ExtraParamSize   if PCM, then doesn't exist
          X   ExtraParams      space for extra parameters

The "data" subchunk contains the size of the data and the actual sound:

36        4   Subchunk2ID      Contains the letters "data"
                               (0x64617461 big-endian form).
40        4   Subchunk2Size    == NumSamples * NumChannels * BitsPerSample/8
                               This is the number of bytes in the data.
                               You can also think of this as the size
                               of the read of the subchunk following this 
                               number.
44        *   Data             The actual sound data.*\
*/

const uint16_t WAVE_FORMAT_PCM = 1;


struct WaveHeader {
  char     riff[4];        // "RIFF"
  uint32_t riffSize;       // fileSize - 8
  char     riffType[4];    // "WAVE"
  char     fmt[4];         // "fmt "
  uint32_t fmtSize;        // 16
  uint16_t audioFormat;    // PCM = 1
  uint16_t nChannel;       // 1
  uint32_t sampleRate;     // 44100, 22050, etc
  uint32_t byteRate;       // sampleRate * nChannel*bitsPerSample/8
  uint16_t blockAlign;     // 2 = nChannels * bitsPerSample/8
  uint16_t bitsPerSample;  // 16 
  char     data[4];        // "data"
  uint32_t dataSize;       // number of data bytes.  fileSize - 44
};
bool writeWaveHeader(SdBaseFile* file, uint32_t rate);
#endif  // SD_FAT_VERSION
#endif  WaveFile_h //