// Record and play 16-bit 44.1 ksps Wave files.
// pins are defined in CrossRoads.h.
#include <CrossRoads.h>
#include <SdFat.h>
#include <SdFatUtil.h>

// Set next line for correct SD chip select.
const uint8_t SD_CS_PIN = SS;

// Write Wave format files if USE_WAVE _FORMAT is nonzero.
#define USE_WAVE_FORMAT 1

// Preallocated file size.
const uint32_t MAX_FILE_SIZE_BLOCKS = 204800;

// Buffer size in 512 byte blocks.
#if defined(__AVR_ATmega1284P__) || defined(__AVR_ATmega1284__)
// Use 8KB of 16 KB for 1284P.
const size_t BUF_BLOCK_COUNT = 16;
#else  // BUF_BLOCK_COUNT
// use 2 KB for Mega or 644P
const size_t BUF_BLOCK_COUNT = 4;
#endif  // BUF_BLOCK_COUNT
//------------------------------------------------------------------------------
//#define DBG_12_BIT_TEST
#ifndef DBG_12_BIT_TEST
AD7680<ADC_CS_PIN, ADC_SDATA_PIN, ADC_SCLK_PIN, USE_WAVE_FORMAT> adc;
AD5662_USART1<DAC_SYNC_PIN, USE_WAVE_FORMAT> dac;
// bit-bang DAC driver
//AD5662<DAC_SYNC_PIN, DAC_DIN_PIN, DAC_SCLK_PIN, USE_WAVE_FORMAT> dac;
#else  // DBG_12_BIT_TEST
MCP3201<ADC_CS_PIN, ADC_SDATA_PIN, ADC_SCLK_PIN, USE_WAVE_FORMAT> adc;
MCP4921_USART1<DAC_SYNC_PIN, USE_WAVE_FORMAT> dac;
// bit-bang DAC driver
//MCP4921<DAC_SYNC_PIN, DAC_DIN_PIN, DAC_SCLK_PIN, USE_WAVE_FORMAT> dac;
#endif  // DBG_12_BIT_TEST

SdFat sd;
SdFile file;

// Buffer size in 16-bit samples.
const size_t BUF_SIZE = 256*BUF_BLOCK_COUNT;

#if USE_WAVE_FORMAT
// Signed ADC/DAC Wave data.
typedef int16_t data_t;
#else  // USE_WAVE_FORMAT
// Unsigned ADC/DAC data.
typedef uint16_t data_t;
#endif  // USE_WAVE_FORMAT
// Sample buffer.
data_t buf[BUF_SIZE];

// Buffer in index.
volatile size_t bufIn = 0;
// Buffer out index.
volatile size_t bufOut = 0;
// Overflow error for ADC.
volatile bool bufOverflow = false;
// Underflow error for DAC.
volatile bool bufUnderflow = false;
//------------------------------------------------------------------------------
// Timer ISR for ADC.
ISR(TIMER1_COMPA_vect) {
  // Use local in so compiler can optimize.
  size_t in = bufIn;
  // Save sample in buffer.
  buf[in++] = adc.read();
  // Check for wrap of circular buffer.
  if (in == BUF_SIZE) in = 0;

  // Check for overflow.
  if (in == bufOut) {
    // Buffer full, can't advance in.
    bufOverflow = true;
  } else {
    // Sample saved in buffer.
    bufIn = in;
  }
}
//------------------------------------------------------------------------------
// Timer ISR for DAC.
ISR(TIMER1_COMPB_vect) {
  size_t out = bufOut;
  if (out != bufIn) {
    dac.write(buf[out++]);
    if (out == BUF_SIZE) out = 0;
    bufOut = out;
  } else {
    bufUnderflow = true;
  }
}
//------------------------------------------------------------------------------
void error(char* msg) {
  Serial.print(F("Error halt: "));
  Serial.println(msg);
  while(1);
}
//------------------------------------------------------------------------------
void playFile() {
  uint32_t minMicros = 0XFFFFFFFF;
  uint32_t maxMicros = 0;
   uint8_t maxFreeSpace = 0;
  char name[13];
  if (!file.isOpen()) {
    Serial.println(F("No current file"));
    return;
  }
  file.rewind();
  int n = file.read(buf, 512*(BUF_BLOCK_COUNT - 1));
  if (n <= 0) error("first read");
  bufIn = n/2;
  bufOut = USE_WAVE_FORMAT ? sizeof(WaveHeader)/sizeof(data_t) : 0;
  bufUnderflow = false;
  file.getFilename(name);
  Serial.print(F("Playing "));
  Serial.println(name);
  while (Serial.read() >= 0);
  Serial.println(F("Type any character to stop."));
  uint32_t t = millis();
  dacStart();
  while (!Serial.available() && !bufUnderflow) {
    // Fetch current input block.  Don't need to disable interrupts.
    uint8_t inBlock = bufIn >> 8;
    uint8_t outBlock = bufOut >> 8;
    uint8_t space;
    // Skip if no space or bufUnderflow.
    if (inBlock == outBlock) continue;
    // Calculate new in.
    size_t in = bufIn + 256;
    if (in == BUF_SIZE) in = 0;
    // Make sure circular buffer will have an empty space.
    if (in == bufOut) continue;

    uint32_t m = micros();
    int n = file.read(&buf[bufIn], 512);
    m = micros() - m;
    if (n == 0) break;
    if (n != 512) error("read");
    // Don't need to disable interrupts since low byte is always zero.
    bufIn = in;
    if (m > maxMicros) maxMicros = m;
    if (m < minMicros) minMicros = m;
    // Refresh out index.
    outBlock = bufOut >> 8;
    if (outBlock < inBlock) {
      space = BUF_BLOCK_COUNT + outBlock - inBlock;
    } else {
      space = outBlock - inBlock;
    }
    if (space > maxFreeSpace) maxFreeSpace = space;
  }
  while (!bufUnderflow) {}
  dacStop();
  t = millis() - t;
  if (file.fileSize() != file.curPosition() && !Serial.available()) {
    Serial.println(F("***Underflow***"));
  }
  Serial.print(F("Min read time: "));
  Serial.print(minMicros);
  Serial.println(F(" usec"));
  Serial.print(F("Max read time: "));
  Serial.print(maxMicros);
  Serial.println(F(" usec"));
  Serial.print(F("Max used buffer: "));
  Serial.print(maxFreeSpace + 1);
  Serial.println(F(" blocks"));
  Serial.print(F("File size: "));
  Serial.print(file.curPosition());
  Serial.println(F(" bytes"));
  Serial.print(F("Play Time: "));
  Serial.print(0.001*t);
  Serial.println(F(" sec"));
  Serial.print(F("Rate: "));
  Serial.print(0.5*file.curPosition()/t);
  Serial.println(F(" ksps"));
  Serial.println(F("Done!"));
}
//------------------------------------------------------------------------------
void recordFile() {
  uint32_t bgnBlock, endBlock;
  uint32_t fileSizeBlocks = 0;
  uint32_t minMicros = 0XFFFFFFFF;
  uint32_t maxMicros = 0;
  uint8_t maxUsedBuf = 0;
  Sd2Card* card = sd.card();
  file.close();

  // create a new binary file
  char name[13];
#if USE_WAVE_FORMAT
  strcpy_P(name, PSTR("WAVE00.WAV"));
#else  // USE_WAVE_FORMAT
  strcpy_P(name, PSTR("DATA00.BIN"));
#endif  // USE_WAVE_FORMAT
  for (uint8_t n = 0; n < 100; n++) {
    name[4] = '0' + n / 10;
    name[5] = '0' + n % 10;
    if (!sd.exists(name)) break;
  }
  // Create large contiguous file.
  if (!file.createContiguous(sd.vwd(), name, 512*MAX_FILE_SIZE_BLOCKS)){
    error("file.create");
  }
  // Get range of blocks.
  file.contiguousRange(&bgnBlock, &endBlock);

  // Erase the file.
  if (!card->erase(bgnBlock, endBlock)) error("erase");

  // Tell card to prepare for a multi-block write.
  if (!card->writeStart(bgnBlock, MAX_FILE_SIZE_BLOCKS)) error("writeStart");

  Serial.print(F("Writing "));
  Serial.println(name);
  while (Serial.read() >= 0);
  Serial.println(F("Type any character to stop."));

  bufOverflow = false;
  bufIn = USE_WAVE_FORMAT ? sizeof(WaveHeader)/sizeof(data_t) : 0;
  bufOut = 0;
  uint32_t t = millis();
  adcStart();
  while (!Serial.available() && !bufOverflow
         && fileSizeBlocks < MAX_FILE_SIZE_BLOCKS) {
    // Fetch current input block.  Don't need to disable interrupts.
    uint8_t inBlock = bufIn >> 8;
    uint8_t outBlock = bufOut >> 8;
    uint8_t used;
    // Skip if no full blocks.
    if (inBlock == outBlock) continue;

    uint32_t m = micros();
    if (!card->writeData((uint8_t*)&buf[bufOut])) error("writeData");
    m = micros() - m;
    if (m > maxMicros) maxMicros = m;
    if (m < minMicros) minMicros = m;
    fileSizeBlocks++;
    size_t out = bufOut + 256;
    if (out == BUF_SIZE) out = 0;
    // Don't need to disable interrupts since low byte is always zero.
    bufOut = out;

    // Refresh in index.
    inBlock = bufIn >> 8;
    // Calculate used number of buffer blocks.
    if (inBlock < outBlock) {
      used = BUF_BLOCK_COUNT + inBlock - outBlock + 1;
    } else {
      used = inBlock - outBlock + 1;
    }
    if (used > maxUsedBuf) maxUsedBuf = used;
  }
  // Stop ADC timer interrupts.
  adcStop();
  // Tell the card we are done streaming data.
  if (!card->writeStop()) error("writeStop");

  t = millis() - t;

  uint32_t used = 512*fileSizeBlocks;
  if (file.fileSize() > used) {
    file.truncate(used);
  }
#if USE_WAVE_FORMAT
  if (!writeWaveHeader(&file, 44100)) error("wave header");
#endif  // USE_WAVE_FORMAT
  file.sync();

  if (bufOverflow) {
    Serial.println(F("**bufOverflow Error**"));
  } else if (fileSizeBlocks == MAX_FILE_SIZE_BLOCKS) {
    Serial.println(F("Max file size reached"));
  }
  Serial.print(F("Min write time: "));
  Serial.print(minMicros);
  Serial.println(F(" usec"));
  Serial.print(F("Max write time: "));
  Serial.print(maxMicros);
  Serial.println(F(" usec"));
  Serial.print(F("Max used buffer: "));
  Serial.print(maxUsedBuf);
  Serial.println(F(" blocks"));
  Serial.print(F("File size: "));
  Serial.print(file.fileSize());
  Serial.println(F(" bytes"));
  Serial.print(F("Record time: "));
  Serial.print(0.001*t);
  Serial.println(F(" Sec"));
  Serial.print(F("Rate: "));
  Serial.print(0.5*file.fileSize()/t);
  Serial.println(F(" ksps"));
  Serial.println(F("Done!"));
}
//------------------------------------------------------------------------------
void setup() {
  Serial.begin(9600);
  Serial.print(F("FreeRam: "));
  Serial.println(FreeRam());
  if (!sd.begin(SD_CS_PIN)) error("sd.begin");
  adc.begin();
  dac.begin();
  timer1Rate(44100);
}
//------------------------------------------------------------------------------
void loop() {
  int c;
  while (Serial.read() >= 0);
  Serial.println(F("Type 'r' to record any other to play last file."));
  while ((c = Serial.read()) < 0);
  if (c != 'r') {
    playFile();
  } else {
    recordFile();
  }
}